\ This is an ANS Forth source file

0 [IF]
Copyright (C) 2003    Ken Staton

This program is part of embedded gForth for the Rabbit2000.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
[THEN]



\ **************************************************************
\ File:         ASM.FS
\               Rabbit 2000 Assembler
\ Author:		Ken Staton
\ Log:		11may03
\
\ * Register usage see PRIMS.FS
\
\ THIS IS NOT A COMPLETE ASSEMBLER FOR THE RABBIT2000
\ IT ONLY ASSEMBLES THE CORE FOR GFORTH

include asm/basic.fs

also  Assembler Definitions

: | ; \ WHAT DOES THIS DO???

: [F]  Forth     ; immediate
: [A]  Assembler ; immediate

hex

\ Rabbit2000 Registers

10000 constant REGIDMASK
20000 constant INDMASK
40000 constant CCMASK

\ (dd) word regs
\ (ss) same
3 constant SSMASK
11000 constant rBC
11001 constant rDE
11002 constant rHL
11003 constant rSP

12000 constant rBC'
12001 constant rDE'
12002 constant rHL'

3 constant XYZMASK
\ (xx) differ in 2
11012 constant rIX
\ (yy) differ in 2
11022 constant rIY
11122 constant rIYo	\ offset (IY+d)
\ (zz) differ in 2,3
11032 constant rzHL
11033 constant rAF

\ byte regs aren't compared with word regs...
7 constant RMASK
10000 constant rB
10001 constant rC
10002 constant rD
10003 constant rE
10004 constant rH
10005 constant rL	\ NOTE NO 6
10007 constant rA

\ condition codes
40000 constant cc=NZ
40001 constant cc=Z
40002 constant cc=NC
40003 constant cc=C
40004 constant cc=LZNV
40005 constant cc=LOV
40006 constant cc=P
40007 constant cc=M


\ Rabbit 2000 Addressing modes

00 constant ()
01 constant (REG)
02 constant (ADR)


\ Rabbit 2000 assembler variables

| Variable size         \ 1 word, 0 byte, -1 undefined
| Variable displaced    \ 1 direct, 0 nothing, -1 displaced
| Variable displacement
| Variable isize        ( specifies Size by prefix )
| Variable mode		( specify mode by prefix )
| Variable ind-mode	( REG, ADR )
| Variable opok

0 mode ! \ be sure it is initialized!

| : Size: ( n -- )  Create c,  Does>  c@ isize ! ;
  0 Size: byte      1 Size: word


| : setsize              isize @  size ! ;
| : moderr               word true Abort" invalid" ;
| : ?moderr ( f -- )     IF moderr THEN ;

| : not ( t|f -- f|t ) -1 xor ; \ negates logical values only
| : ?word   ( n -- t|f ) FF > ;
| : ?byte   ( n -- t|f ) FF <= ;
| : ?reg ( n -- t|f ) dup REGIDMASK >= swap CCMASK < and ;
| : ?immed ( -- t|f ) mode @ () = if true else false then ;
| : ?) ( -- t|f ) mode @ () = not if true else false then ;
| : op-ok true opok ! ;
| : op-not-ok false opok ! ;
| : ?opok opok @ ;
| : ?cc ( n -- t|f ) 40000 40003 within ;
| : ?rel ( n -- t|f ) -128 127 within ;


\ modifies the IND bit of the regID
\ and sets ind-mode and sets mode (is all of this really necessary?)
: ) ( id -- id2 ) dup ?reg 
  if INDMASK or (REG) ind-mode ! mode @ (REG) or mode !
  else (ADR) ind-mode ! mode @ (ADR) or mode !
  endif ;

\ indirect register IDs

: (rHL) rHL INDMASK or ;
: (rIX) rIX INDMASK or ;
: (rIY) rIY INDMASK or ;

\ test register select grouping memberships

\ ?member is used in ?r~~ t|f is set false before 1st use
\ then set true if a member matches
| : ?member ( rID1 t|f rID -- )
	2 pick = if drop true endif ;

| : ?rSS ( regID1 -- t|f )
	false
	rBC ?member
	rDE ?member
	rHL ?member
	rSP ?member
	nip
;

' ?rSS ALIAS ?rDD

| : ?rDD'
	false
	rBC' ?member
	rDE' ?member
	rHL' ?member
	nip
;

| : ?rZZ 
	false
	rBC ?member
	rDE ?member
	rHL ?member
	rAF ?member
	nip
;

| : ?rYY
	false
	rBC ?member
	rDE ?member
	rIY ?member
	rSP ?member
	nip
;

| : ?rXX
	false
	rBC ?member
	rDE ?member
	rIX ?member
	rSP ?member
;

\ used in logic opcodes...
| : ?rG
	false
	rB ?member
	rC ?member
	rD ?member
	rE ?member
	rH ?member
	rL ?member
	rA ?member
	nip
;

' ?rG ALIAS ?rR

\ init-next-op must be done AFTER opcode has been processed
\ not before because mode is set AS line is read

| : init-next-op 0 mode !  ;


: loadSP ED X c, 7B X c, X , ; \ address to read was on stack
: storSP ED X c, 73 X c, X , ; \ address to write was on stack

: ex(SP),HL ED X c, 54 X c, ;

: 1mi create c, does> c@ X c, ;
: 2mi create c, c, does> dup c@ swap 1 + c@ X c, X c, ;
: 1miop8 create c, does> c@ X c, X c, ;
: 1miop16 create c, does> c@ X c, X , ;
: logic create c, does> c@  
	over dup REGIDMASK xor swap ?rG not ( -- t|f ) Abort" not logic register"
	+ X c, drop ;

\ byte reg parameter... e.g. rD and, is A = A & rD
a0 logic andr,
a8 logic xorr,
b0 logic orr,

: or, 
op-not-ok
depth 2 >= 
if \ be sure there are 2 things to test before trying...
  over rDE = 
  if \ src must be DE (not an error if other)
    dup 
    rHL = if %11101100 X c, op-ok endif
    dup
    rIX = if %11011101 X c, %11101100 X c, op-ok endif
    dup
    rIY = if %11111101 X c, %11101100 X c, op-ok endif

   ?opok
   if 2drop
   else true abort" ERROR: bad dst in rDE r?? or,"
   endif
  endif
endif  

dup ?rR 
if 		\ byte reg only
  orr,
endif
;

: xor, xorr, ; \ byte reg only for now

: and,
op-not-ok
depth 2 >= 
if \ be sure there are 2 things to test before trying...
  over rDE = 
  if \ src must be DE (not an error if other)
    dup 
    rHL = if %11011100 X c, op-ok endif
    dup
    rIX = if %11011101 X c, %11011100 X c, op-ok endif
    dup
    rIY = if %11111101 X c, %11011100 X c, op-ok endif

   ?opok
   if 2drop
   else true abort" ERROR: bad dst in rDE r?? and,"
   endif
  endif
endif  

dup ?rR 
if 		\ byte reg only
  andr,
else
  dup ?byte if %11100110 X c, X c, op-ok endif
endif
;

: add, ( src dst -- )
op-not-ok

dup
rHL = 
if over ?rSS
	if over SSMASK and 4 lshift %00001001 or X c, op-ok 
	else true abort" ERROR bad src in r?? rHL add," 
	endif
endif

dup
rIY =
if over ?rYY
	if %11111101 X c, over XYZMASK and 4 lshift %00001001 or X c, op-ok
	else true abort" ERROR bad src in r?? rIY add,"
	endif
endif

?opok
if 2drop
else true abort" ERROR: problem with add,"
endif
;


\ BIT b,r r:r{B,C,D,E,H,L,A}  bit# reg BIT

: bit, ( reg bit# -- )
	over ?rR not if true abort" ERROR bit, bad register!" endif
	cb X c, 3 lshift $40 + REGIDMASK xor + X c, ;

: bool, ( reg -- )
	op-not-ok
	dup
	rHL = if %11001100 X c, op-ok then
	dup
	rIX = if %11011101 X c, %11001100 X c, op-ok then
	dup
	rIY = if %11111101 X c, %11001100 X c, op-ok then

	?opok
	if drop init-next-op
	else
	  true abort" ERROR: bad register for bool,"
	endif
;


: addHL, ( reg -- )	\ dest always HL = HL + rSS
	op-not-ok
	dup
	rDE = if 19 X c, op-ok then
	dup
	rHL = if 29 X c, op-ok then
	dup
	rSP = if 39 X c, op-ok then 

	?opok
	if drop init-next-op
	else
	  true abort" ERROR: bad register for addHL "
	endif
;


: pop, ( reg -- ) 
	op-not-ok
	dup
	rIX = if dd X c, e1 X c, op-ok then
	dup
	rIY = if fd X c, e1 X c, op-ok then
	dup
	rBC = if c1 X c, op-ok then
	dup
	rDE = if d1 X c, op-ok then
	dup
	rHL = if e1 X c, op-ok then
	dup
	rAF = if f1 X c, op-ok then 

	?opok 
	if drop init-next-op 
	else
	  true abort" ERROR: bad register for POP "
	endif
;


: push, ( reg -- )
	op-not-ok
	dup
	rIX = if dd X c, e5 X c, op-ok then
	dup
	rIY = if fd X c, e5 X c, op-ok then
	dup
	rBC = if c5 X c, op-ok then
	dup 
	rDE = if D5 X c, op-ok then
	dup
	rHL = if e5 X c, op-ok then
	dup
	rAF = if f5 X c, op-ok then

	?opok
	if drop init-next-op
	else
	  true abort" ERROR: bad register for PUSH "
	endif
;


: dec, ( reg -- ) 
	op-not-ok
\	dup
\	rHL = if 2b X c, op-ok endif
	dup
	?rSS if dup SSMASK and 4 lshift %00001011 or X c, op-ok endif 
	dup
	rIX = if %11011101 X c, %00101011 X c, op-ok endif
	dup 
	rIY = if %11111101 X c, %00101011 X c, op-ok endif
	dup
	rA = if 3d X c, op-ok endif
	?) if
		dup
		rHL = if 35 X c, op-ok endif
	    endif

	?opok
	if drop init-next-op
	else
	  true abort" ERROR: bad register for DEC "
	endif
;


: inc, ( reg -- )
	op-not-ok
	dup
	rBC = if 03 X c, op-ok then
	dup
	rDE = if 13 X c, op-ok then
	dup
	rHL = if 23 X c, op-ok then
	dup
	rSP = if 33 X c, op-ok then
	dup
	rIX = if %11011101 X c, %00100011 X c, op-ok endif
	dup
	rIY = if %11111101 X c, %00100011 X c, op-ok endif

	dup
	?rR if dup REGIDMASK xor %00000100 or X c, op-ok endif

	?opok
	if drop init-next-op
	else
	  true abort" ERROR: bad register for INC "
	endif
;

: ex, ( rHL rDE -- )
op-not-ok
over
rHL = not if true abort" ERROR in rHL rDE|rDE' ex, " endif

dup
rDE = if %11101011 X c, op-ok endif

dup
rDE' = if %11100011 X c, op-ok endif

?opok
if 2drop init-next-op
else
  true abort" ERROR: problem in rHL rDE|rDE' ex, " 
endif
;

: sub, ( src -- )
op-not-ok
dup
?rR if dup REGIDMASK xor %10010000 or X c, op-ok endif

?opok
if drop init-next-op
else
  true abort" ERROR: problem in rR sub,"
endif
;


: subc, ( src dst -- )
op-not-ok
dup
rHL = if over ?rSS 
	if over SSMASK and 4 lshift %11101101 X c, %01000010 or X c, op-ok
	else true abort" ERROR: src not rSS in subc," 
	endif
      endif
?opok
if 2drop init-next-op
else
  true abort" ERROR: problem in subc,"
endif
;

: rr, ( reg -- )
op-not-ok

dup
rHL = if %11111100 X c, op-ok endif

dup
?rR if %11001011 X c, dup REGIDMASK xor %00011000 or X c, op-ok endif

?opok
if drop init-next-op
else
  true abort" ERROR: rr, "
endif
;

: rl, ( reg -- )
op-not-ok

dup
?rR if %11001011 X c, dup REGIDMASK xor %00010000 or X c, op-ok endif

?opok
if drop init-next-op
else
  true abort" ERROR rl, "
endif
;


\ std (manufacturer's) syntax is: LD dest,src
\ forth syntax is: src mode dest mode LD

\  1 indirect destination
\ case	dest	src		syntax
\  1a (BC)(DE)	A		rA r?? ) ld
\  1b	(HL)	HL/r		rR rHL ) ld
\  1c	(HL)	n8		n rHL ) ld
\  1d	(mn)	A/HL/IX/IY/ss	r?? mn ) ld
\  1e	(IX+d)	HL/r		rHL/r rIX ) ld  d=0 always!
\  1f	(IY+d)	HL/r		rHL/r rIY ) ld  d=0 always!


\  2 indirect source
\ case	dest	src		syntax
\  2a	A 	(BC)|(DE)	r?? ) rA ld
\  2b	A	(mn)		mn ) rA ld
\  2c	dd 	(mn)		mn ) rDD ld
\  2d	HL 	(mn)		mn ) rHL ld
\  2d	HL	(HL/IX/IY+d)	r?? ) rHL ld  d=0 always!
\  2e	r	(HL/IX/IY+D)	r?? ) r ld  d=0 always!


\  3 direct
\ case	dest		src		syntax
\  3a	dd/IX/IY 	mn16		mn rDD ld
\  3b	HL		IX/IY		r?? rHL ld
\  3bb	IX/IY		HL		rHL r?? ld
\  3c	r 		g		rG rR ld
\  3d	r 		n8		n rR ld
\  3e	SP 		HL/IX/IY	r?? rSP ld
\  3f	dd'		BC/DE		r??' r?? ld
\ others too!

\ *** LD CASES *** *** LD CASES *** *** LD CASES *** *** LD CASES ***

\ src mode dest mode LD
\ immediate is implicit. all words that use mode variable
\ must clear it once opcodes are generated.
\ tell if mn ) using mode _and_ the src or dst redID (no IND bit)
\ tell which register in 2 reg operand instructions using regID

\ immediate data loads
: case3ad ( #data reg -- )
op-not-ok
dup
?rDD  \ treat as word if rDD
if \ case 3a
  dup REGIDMASK xor 4 lshift %00000001 or X c, over X , op-ok endif

dup
?rR  \ treat as byte if rR
if  \ case 3d
  over ?word if true abort" ERROR: word data doesnt'fit in byte reg!" endif
  dup REGIDMASK xor 3 lshift %00000110 or X c, over X c, op-ok endif

dup
rIX = if %11011101 X c, %00100001 X c, over X , op-ok endif

dup
rIY = if %11111101 X c, %00100001 X c, over X , op-ok endif
	
?opok
if 2drop init-next-op
else
  true abort" ERROR: bad register for LD "
endif
;



\ register direct loads
: case3c ( rS rD -- )
REGIDMASK xor 3 lshift swap REGIDMASK - or 40 or X c,
;

: case3b ( src rHL -- )
drop \ already verified dst to get here
 
\ only valid register direct are from IX, IY
op-not-ok
dup 
rIX = if %11011101 X c, %01111100 X c, op-ok endif
dup 
rIY = if %11111101 X c, %01111100 X c, op-ok endif 

?opok
if drop init-next-op
else
  true abort" ERROR: bad register direct to HL"
endif
;

: case3bb ( rHL dst -- )
\ only do reg direct to IX, IY
op-not-ok
dup
rIX = if %11011101 X c, %01111101 X c, op-ok endif
dup
rIY = if %11111101 X c, %01111101 X c, op-ok endif

?opok
if 2drop init-next-op
else 
  true abort" ERROR: bad destination register direct from HL"
endif
;

: case3e ( src rSP -- )
\ only valid direct are from HL, IX, IY
op-not-ok
drop \ already verified dst to get here
dup
rHL = if %11111001 X c, op-ok endif
dup 
rIX = if %11011101 X c, %11111001 X c, op-ok endif
dup 
rIY = if %11111101 X c, %11111001 X c, op-ok endif 

?opok
if drop init-next-op
else
  true abort" ERROR: bad register direct to HL"
endif
;

: case3f ( BC|DE rDD' -- )
op-not-ok
dup ?rDD' not if true abort" ERROR ld case3f dst rDD' " endif

over
rBC = if dup REGIDMASK xor 4 lshift 
	%11101101 X c, %01001001 or X c, op-ok
      endif

over
rDE = if dup REGIDMASK xor 4 lshift
	%11101101 X c, %01000001 or X c, op-ok
      endif

?opok
if 2drop init-next-op
else
  true abort" ERROR: ld case3f"
endif
;

: case2ab ( src rA -- )
drop \ already verified dst to get here
op-not-ok
dup
?reg not 
if %00111010 X c, X , op-ok 
else
  dup 
  rBC INDMASK + = if drop %00001010 X c, op-ok endif
  dup
  rDE INDMASK + = if drop %00011010 X c, op-ok endif
  dup
  rHL INDMASK + = if drop %01111110 X c, op-ok endif \ case2e dst=a
endif

?opok
if init-next-op
else
  true abort" ERROR: problem xx ) rA ld"
endif
;

: case2c ( src rDD -- )
op-not-ok

over REGIDMASK and 0<> 
if true abort" ERROR: ld, case2c src must be number"
endif

dup
?rDD
if
  %11101101 X c,
  SSMASK and 4 lshift %01001011 or X c,
  X , op-ok
else
  true abort" ERROR: mn) rDD ld (case2c)"
endif

?opok
if init-next-op
else
  true abort" ERROR: impossible  mn ) rDD ld"
endif
;

: case2d ( src rHL -- )
drop \ verified dst to get here
op-not-ok

dup
?reg not
if dup %00101010 X c, X , op-ok endif

dup
rHL INDMASK or =
if %11011101 X c, %11100100 X c, 0 X c, op-ok endif

dup
rIX INDMASK or =
if %11100100 X c, 0 X c, op-ok endif

dup
rIY INDMASK or =
if %11111101 X c, %11100100 X c, 0 X c, op-ok endif

dup \ IY+d ( d -1 -1 src  -- )
rIYo INDMASK or = 
if 
  3 pick \ get d
  \ FIXME range check
  %11111101 X c, %11100100 X c, X c, op-ok 
  swap drop \ get rid of one of the dummy stack items here
endif

?opok
if drop init-next-op
else true abort" ERROR: problem src ) rHL ld"
endif
;

: case2e ( src.i rR -- )
op-not-ok
over
(rHL) =
  if dup REGIDMASK xor 3 lshift %01000110 or X c, op-ok endif

over
(rIX) = 
  if 
    dup REGIDMASK xor 3 lshift %11011101 X c, 
    %01000110 or X c, 0 X c, op-ok 
  endif

over
(rIY) =
  if 
    dup REGIDMASK xor 3 lshift %11111101 X c, 
    %01000110 or X c, 0 X c, op-ok
  endif

?opok
if 2drop init-next-op
else
  true abort" ERROR: problem in  r?? ) rR ld"
endif
;

: case1a ( rA rBCi|rDEi -- )
swap
rA = not if true abort" ERROR: src not A in A BC ) LD" endif

op-not-ok
dup
rBC INDMASK + = if %00000010 X c, op-ok endif
dup
rDE INDMASK + = if %00010010 X c, op-ok endif

?opok
if drop init-next-op
else
  true abort" ERROR: bad register direct to HL"
endif
;

: case1bc ( rHL|rR|n rHLi -- )
drop \ dst HL ) to get here
op-not-ok

dup
?reg not 
if dup ?word not
  if dup %00110110 X c, X c, op-ok  
  else true abort" ERROR: n8 HL ) ld larger than byte!" 
  endif
else \ it is a register...
  dup ?rR
  if dup %01110000 or X c, op-ok endif
  dup rHL = if %11011101 X c, %11110100 X c, 0 X c, op-ok endif
endif

?opok
if drop init-next-op
else
  true abort" ERROR: problem in xx HL ) ld"
endif
;

: case1d ( src mni  -- )

op-not-ok
over
rA = if %00110010 X c, dup X , op-ok endif
over
rHL = if %00100010 X c, dup X , op-ok endif
over
rIX = if %11011101 X c, %00100010 X c, dup X , op-ok endif
over
rIY = if %11111101 X c, %00100010 X c, dup X , op-ok endif
over ?rSS 
  if %11101101 X c, 
    over SSMASK and 4 lshift %01000011 or X c, 
    dup X , op-ok
  endif

?opok
if 2drop init-next-op
else
  true abort" ERROR: problem in rSS mn ) ld"
endif
;

: case1e ( src rIX -- )
\ dst = IX ) to get here

op-not-ok
over
rHL = if %11110100 X c, 0 X c, op-ok endif
over
?rR if %11011101 X c, over REGIDMASK xor %01110000 or X c, 0 X c, op-ok endif

?opok
if 2drop init-next-op
else 
  true abort" ERROR: problem in rHL/rR  rIX ) ld,"
endif
;

: case1f ( src rIY -- )
\ dst = IY ) to get here

op-not-ok
over
rHL = if %11111101 X c, %11110100 X c, 0 X c, op-ok endif
over
?rR if %11111101 X c, over REGIDMASK xor %01110000 or X c, 0 X c, op-ok endif

?opok
if 2drop init-next-op
else 
  true abort" ERROR: problem in rHL/rR  rIY ) ld,"
endif
;
 
\ LD is a mess...
\ figure out which case we've got...
\ each of these tests looks at either the src or dst
\ sometimes at both src & dst, but when a case is found
\ and processed, the stack parms are consumed!!!
\ so put dummy items (pair) on stack to test until
\ done considering all cases...
\ MUST do non-register (number) cases first!

: ld, ( src dst -- )
-1 -1 2swap
\ consider case 3 1st:
\ no ind -> case 3
 mode @ () =
 if \ no indirect
   over ?reg not  \ direct with number (which must be src)
	if \ src is not a register
	  case3ad \ cases 3(a,d)
	endif
   over rBC = if dup ?rDD' if case3f endif endif
   over rDE = if dup ?rDD' if case3f endif endif
   dup rHL = if case3b endif \ dst is HL
   over rHL =
	if \ src is HL
	  dup rIX = if case3bb endif
	  dup rIY = if case3bb endif
	endif
   dup rSP = if case3e endif \ dst is SP
   over ?rG if dup ?rR if case3c endif endif
 else \ indirect... find it!
   over ?reg 
   if \ src is register 
     over INDMASK and 0<> 
     if \ and is indirect
       true 
     else \ is not indirect, so the dst must be
       false
     endif
   else \ src is number
     dup INDMASK and 0<> 
     if \ and dst is indirect
       false 
     else \ dst is not indirect, so the src must be
       true
     endif
   endif ( src dst f -- ) \ t = (src) | f = (dst)
   if \ consider case 2 (src) next:
     dup rA = if case2ab endif \ dst is A
     dup rHL = if case2d endif \ dst is HL
     dup ?rR if case2e endif \ dst is set R
     dup ?rDD if case2c endif \ dst is set DD
   else \ consider case 1 (dst) last:
     dup ?reg not if case1d endif \ dst is not a register (mn)	
     dup INDMASK xor rBC = if case1a endif \ dst is (BC)
     dup INDMASK xor rDE = if case1a endif \ dst is (DE)
     dup INDMASK xor rHL = if case1bc endif \ dst is (HL)
     dup INDMASK xor rIX = if case1e endif \ dst is (IX)
     dup INDMASK xor rIY = if case1f endif \ dst is (IY)
   endif
 endif \ indirect
2drop \ lose the dummy stack values
;

\ *** END LD *** *** END LD *** *** END LD *** *** END LD *** 

\ must distinguish  'jp mn'  from 'jp (HL)'
\ mode = direct | indirect
\ mn jp,
\ rHL ) jp,
\ mn CC jp, 
: jp,  \ ( addr |  rNN -- ) # addr jp, | r?? ) jp,
	op-not-ok
	
	?immed	
	if 
	  dup ?cc not 
	  if dup C3 X c, X , op-ok 
	  else CCMASK xor 3 lshift %11000010 or X c, dup X , op-ok
	  endif
	else 
	  dup
	  (rHL) = if E9 X c, op-ok then
	  dup
	  (rIX) = if DD X c, E9 X c, op-ok then
	  dup
	  (rIY) = if FD X c, E9 X c, op-ok then
	endif
	?opok
	if drop init-next-op
	else
	  true abort" ERROR: bad register for JP () "
	endif
;


: jr, ( addr cc -- )  \ only conditional for now...
swap
X here 2 + - dup ?rel not 
if true abort" ERROR: jr offset out of range! " endif
( cc rel-addr -- )
over ?cc 
if swap CCMASK xor 3 lshift %00100000 or X c, X c, endif
;


: call, ( addr -- )
%11001101 X c, X , 
;


\ used for debug with a binary image editor...
| : ken 20 X c, 20 X c, 4b X c, 65 X c, 6e X c, 20 X c, 20 X c, ;

\ all single opcode, no operand instructions:

%11011011 1mi ioe,
%11010011 1mi ioi,
%00000000 1mi nop,
%11001001 1mi ret,
%01110110 1mi altd,
%00111111 1mi ccf,
%00101111 1mi cpl,
%11011001 1mi exx,
%11110111 1mi mul,
%00010111 1mi rla,
%00000111 1mi rlca,
%00011111 1mi rra,
%00001111 1mi rrca,
%00110111 1mi scf,

\ all double opcode,no operand instructions:

%11101101 %01011101 2mi ipres,
%11101101 %10101000 2mi ldd,
%11101101 %10111000 2mi lddr,
%11101101 %10100000 2mi ldi,
%11101101 %10110000 2mi ldir,
%11101101 %01000101 2mi lret,
%11101101 %01000100 2mi neg,
%11101101 %01001101 2mi reti,

\ opcodes for conditional relative jump:
\ the offset is placed by the high level construct
\ -128..+127  not far enough? factor!

%00100000 ( NZ ) 	constant 0=
%00101000 ( Z )		constant 0<>
%00110000 ( NC )	constant u<
%00111000 ( C )		constant u>=

\ IF, compiles the conditional branch opcode from tos 
\ as a word (the high byte is zero)
\ then puts here-1 on tos (points to the zero high byte)

  : IF,          X ,   X here  1- ;

\ THEN, gets here & adds 1 to the reserved cell on tos
\ and subtracts them to get the offset
\ the offset is checked for valid relative value range
\ and stored at the reserved cell from IF,

  : THEN,        X here  over 1+ - ( ?range ) swap X c!  ;

\ ELSE, is an "IF," with an unconditional branch opcode
\ and uses "THEN," to complete the IF block

  : ELSE,       %00011000  IF, swap THEN, ;

\ just doing what we need for the core rather than making
\ a full cross assembler...
\  : WHILE,      IF, swap ;
\  : BEGIN,       X here  ;
\  : UNTIL,       X c,   X here  1+ - ( ?range )  X c,  ;
\  : AGAIN,      $eb UNTIL, ;
\  : REPEAT,     AGAIN, THEN, ;


\ (Code)-R2K   (End-Code)-R2K

  : (Code)-R2K
    (code)-1 ;          ' (Code)-R2K IS (code)

  : (End-Code)-R2K
    (end-code)-1 ;      ' (End-Code)-R2K IS (end-code)

