HEX


\ file ( -- )
\ compiles a file a line at a time - emits '.' when ready for next
\ uses memory below return stack as line buffer
\ send ESC key at EOF

: file 
\ rp@ 800 - \ line buffer below return stack
8000 
dup 100 0 fill ( addr -- ) 
>r \ don't use data stack. causes probs when multiline defines
begin
  r@ ( addr -- )
  begin
    key 	\ waits for char ( addr ch -- )
    dup 1B = if 2drop r> drop exit endif
    dup 0D =
    if
      over c!	\ put key in buffer
      1+	\ inc addr
      true
    else
      over c!	\ put key in buffer
      1+	\ inc addr
      false
    endif	( addr flag -- )
  until 	\ 0D at EOL
  r@ ( addr1 addr -- ) - 	\ size
  r@ swap ( addr size -- ) 
  evaluate 
  r@ 100 0 fill 
  [char] . emit
  false		\ exits
until		\ ESC at EOF
;

8000 constant inbuf 	\ Put input buffer at 0x8000
			\ rp@ ~ 0xAE00, so buffer size is limited
			\ to about 0x2000!


\ load ( -- )
\ loads an entire file into a buffer.
\ useful when your terminal emulator doesn't support line based flow control

: load
 inbuf 2000 0 fill 	\ clear buffer
 inbuf 			\ load buffer
 begin 
   key dup 1a ( ^Z ) = 
   if 
     drop 	\ key
     drop 	\ addr (inbuf)
     true 
   else 
     over c! 	\ put key in buffer
     1+ 	\ inc addr
     false  
   endif
 until ;


variable doneL
variable doneB

\ read ( -- )
\ read the source in the buffer after 'load'

: read 
false doneL ! 
false doneB !
0   \ initial index is start of buffer
begin  \ '\' comments whitespaced
   ( index -- )
   dup inbuf + c@ dup ( index ch ch -- )
   0 = 
   if 
     true doneB ! 
     drop \ ch
     1 - ( index-1 -- )
   else 
     ( index ch -- )
     [char] \ = ( index flag -- ) 
     if  \ eol comment
       dup inbuf + 20 swap c! 1+ \ write over '\' with blank

       ( index -- )
       begin \ write over rest of line with blank
         dup inbuf + c@ ( index ch -- ) \ get next char
         dup 0D = swap 0 = or \  is next char 'cr' or null ?
         if true doneL ! 1+
         else dup inbuf + 20 swap c! 1+ 
         endif 
         doneL @ 
       until \ end of line or empty

       false doneL ! \ reset line flag
       [char] . emit \ indicated eol comment line
     else \ no eol comment 
       dup inbuf + c@ \ get current char
       0D = if [char] . emit endif \ indicate no eol comment line
       1+ \ inc index
     endif \ eol comment
   endif \ zero
   doneB @ 
 until \ '\' comments whitespaced
 ( index -- )
 ." Length = " .s
 inbuf swap 
 evaluate
;
  
