\ This is an ANS Forth source file

0 [IF]
Copyright (C) 2003    Ken Staton

This program is part of embedded gForth for the Rabbit2000.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
[THEN]



\ **************************************************************
\ File:         PRIMS.FS
\                 Lowlevel routines for GFORTH on Rabbit2000
\ Author:	Ken Staton
\ Log:		11may03
\		20may03 cpu init code
\		21may03 core coding
\		21jun03 um* code
\		22jun03 cycle improvements : words with next-w,
\
\ * Memory ( use only one 64K-Page ):
\    $0000-$BC00 : program		KEEP OUT OF RABBIT BIOS SPACE!
\    $B400-$B600 : returnstack		builds down
\    $B600-$B800 : datastack		builds down
\    $B820-$BC00 : TIB			TIB $20 above SP! grows up
\ **************************************************************


start-macros

\ R2K Forth register definitions

\ IY is used as a scratch register
\ A is used as a scratch register

\ alternate register use:
\ HL' in um*


\ as word registers:
  ' rSP Alias fsp
  ' rIX Alias frp
  ' rDE Alias fip
  ' rHL Alias W
  ' rBC Alias tos
  ' rIY Alias XW	\ X is already used by cross
\ as byte registers:
  ' rL  Alias WL
  ' rH  Alias WH
  ' rC  Alias tosL
  ' rB  Alias tosH
  ' rE  Alias fipL
  ' rD  Alias fipH
 

\ R2000 System Registers
$A0 constant TACSR	\ timer A
$A4 constant TACR
$AD constant TAT6R
$AB constant TAT5R
$A9 constant TAT4R

$70 constant PEDR	\ port E
$75 constant PEFR
$77 constant PEDDR
$74 constant PECR

$55 constant PCFR	\ port C
$50 constant PCDR

$03 constant STATREG	\ serial port offsets
$04 constant CTLREG

$E0 constant SPORTC
$E3 constant SCSR
$E4 constant SCCR
$E0 constant SCDR

$D0 constant SPORTB
$D3 constant SBSR
$D4 constant SBCR
$D0 constant SBDR

$C0 constant SPORTA
$C3 constant SASR
$C4 constant SACR
$C0 constant SADR

$C0 constant SERIAL	\ select serial port A

$0F constant GCDR

$08 constant WDTCR	\ watchdog
$09 constant WDTTR

$01 constant RTCCR
$02 constant RTC0R
$03 constant RTC1R
$04 constant RTC2R
$05 constant RTC3R
$06 constant RTC4R
$07 constant RTC5R

$00 constant GCRS



\ system dependent macros

\ note that next, leaves the fip value (the CFA) in XW (i.e. IY)
\ this is useful in other core words that use fip.
\ ITC

\ in-line coded for speed (saves jp cycles)...
.( inline next)

  : next-w,		\ next word in W *ci*
	W ) fipL ld,	\ get low byte of addr @ fip 	1b5c $5e
	W inc,		\ inc ip to point to high byte	1b2c $23
	W ) fipH ld,	\ get high byte of addr @ fip	1b5c $56
	W inc,		\ inc ip to point to next addr ( low byte ) 1b2c $23
	W fip ex,	\				1b2c $eb
			\ W = CFA for next word		
	W XW ld,	\ save CFA of next word in XW	2b4c $fd7d
	W ) W ld,	\ ITC get machine code addr	3b11c $dde400
	W ) jp,		\ jmp there			1b4c $e9
  ; 			\ 11 bytes 35 cycles

  : next, 		\ jump to next word
			\ (fip) -> W	fip+
	W fip ex,	\				1b2c $eb
	next-w,
  ;			\ 12 bytes 37 cycles


\ W->tos, macro: 2 cycles (2c) each. 4c total.
\ alternate is W pop, which is 7c

  : W->tos,
	WL tosL ld,
	WH tosH ld,
  ;

\ tos->W, macro: 4c total.
\ alternate is W push, which is 10c

  : tos->W,
	tosL WL ld,
	tosH WH ld,
  ;

  : W->fip,
	WL fipL ld,
	WH fipH ld,
  ;

  : fip->W,
	fipL WL ld,
	fipH WH ld,	
  ;

\ halt, is used for debugging. divide and conquer.

  : halt,
     %00011000 X c, %11111110 X c,		\ halt for now
  ;

end-macros

  unlock
    $0000 $C000 ( addr len -- ) region dictionary
    setup-target
  lock


\ ==============================================================
\ rom starts with jump to GFORTH-kernel (must be at $0000 !!!)
\ ==============================================================
  Label into-forth
     $ffff JP,			\ jump to init code (patched)
				\ then then jump addr=3 (into-forth-init)

  Label into-forth-init 
     $ffff fip LD,            \ ip (patched) (addr 4,5 are overwritten later)
     $B800 fsp LD,            \ sp top \ builds down
     $B600 frp LD,            \ rp top \ builds down
    next,
  End-Label

\ leave room for scratch ram here:
\ these locations are used to return
\ to the rabbit bios and to specify the
\ serial port and baud rate

  Label baud
	nop,
	nop,
  End-Label

  Label port
	nop,
	nop,
  End-Label

  Label save_sp
	nop,
	nop,
  End-Label

  Label msg1
	$0D c, $0A c,
	$0D c, $0A c,
	$52 c, $61 c, $62 c, $62 c, $69 c, $74 c, $20 c, $4F c, $4B c,
	$0D c, $0A c,
	$42 c, $6F c, $6F c, $74 c, $69 c, $6E c, $67 c,
	$20 c, $46 c, $6F c, $72 c, $74 c, $68 c, $2E c, $2E c, $2E c,
	$0D c, $0A c,
	$00 c, $00 c,
  End-Label


  Label (aout) \ (emit) with char in rA
    rAF push,
  Label  aowait
    ioi,
    SERIAL STATREG + ) rA ld,
    rA 2 bit,
    aowait cc=NZ jr,
    rAF pop,
    ioi,
    rA SERIAL ) ld,
    ret,
  End-Label



\ rabbit2000 init...

  Label system-init
	rBC push,
	rDE push,
	rHL push,
	rAF push,
	rIY push,
	rIX push,		\ save state on dyn-c stack

	rSP save_sp ) ld,	\ remember where dyn-c stack was
	rL rA ld,
	rA baud ) ld,		\ baud div is parameter
	rH rA ld,
	rA port ) ld,		\ serial port is parameter

	$53 rA ld,
	ioi,
	rA WDTTR ) ld,

	$54 rA ld,		\ no watchdog!
	ioi,
	rA WDTTR ) ld,

	$44 rA ld,		\ select txa & txc alt-c i/o func
	ioi,
	rA PCFR ) ld,
	
	$0 rA ld,		\ serial port ctl: portc, 8bit async, no int
	ioi,
	rA SERIAL CTLREG + ) ld,

	baud ) rA ld,		\ timer a stuff

	ioi,			\ set all baud divisors...
	rA TAT6R ) ld,
	nop,

	ioi,
	rA TAT5R ) ld,
	nop,

	ioi,
	rA TAT4R ) ld,
	nop,

	$00 rA ld,
	ioi,
	rA TACR ) ld,

	$01 rA ld,
	ioi,
	rA TACSR ) ld,		\ enable timer A

	$00 rA ld,
	ioi,
	rA PECR ) ld,

	$00 rA ld,
	ioi,
	rA PEFR ) ld,

	$80 rA ld,
	ioi,
	rA PEDDR ) ld,		\ port e stuff

	\ delay for initialization!

	$10 rA ld,
Label L006
	$0000 rDE ld,
	$0000 rHL ld,
Label L005
	rHL inc,
	rDE rHL or,
	L005 cc=NZ jr,
	rA dec,
	L006 cc=NZ jr,

	\ boot message...

	msg1 rHL ld,
Label L003
	rHL ) rC ld,
	rHL inc,
Label L004
	ioi,
	SERIAL STATREG + ) rA ld,

	rA 2 bit,
	L004 cc=NZ jr,
	rC rA ld,

	ioi,
	rA SERIAL ) ld,
	nop,

	rHL ) rA ld,
	rA or,
	L003 cc=NZ jr,

\ Label LIL
\	LIL jp,

	into-forth-init JP,	\ jump to forth init
  End-Label


0 [IF]
Start-Macros
  : dout, ( c -- )
    rA ld,
    (aout) call,
  ;
End-Macros
[ELSE]
Start-Macros
  : dout, drop ;	\ nothing added to image
End-Macros
[ENDIF]



\ ==============================================================
\ GFORTH minimal primitive set
\ ==============================================================


\ inner interpreter

\ headerless...
  Code: :docol
    ': dout,
    frp dec,		\ -frp
    frp dec,		\ 		2b4c
    W fip ex,		\ fip->(frp) 	1b2c
    W frp ) ld,		\ 		3b11c
    W fip ex,		\ 		1b2c
    XW W ld,		\ current CFA restored to W 2b4c
    W inc,		\ offset to PFA
    W inc,
    W inc,
    W inc,
    next-w,
  End-Code

  Code: :dovar
   '2 dout,
    tos push,		\ push the data stack (make room for value)
    XW W ld,		\ current CFA restored to W
    W inc,
    W inc,
    W inc,
    W inc,
    W->tos,		\ put addr on tos
    next,
  End-Code

  Code: :dodoes  ( -- pfa ) \ get pfa and execute DOES> part
   '6 dout,
    frp dec,		\ save fip on return stack a byte at a time
    fipH frp ) ld,	\ high byte 1st to keep little endian
    frp dec,
    fipL frp ) ld,
    tos push,		\ make room
    XW W ld,		\ current CFA restored to W
    W inc,
    W inc,
    W inc,
    W inc,
    W->tos,		\ copy PFA to tos
    W dec,
    W dec,
    W ) W ld,		\ get does> addr
    next-w,             \ execute does> part
   End-Code


\ program flow

  Code ;s       ( -- ) \ exit colon definition
   '; dout,
    frp ) fipL ld,	\ get ip off return stack a byte at a time
    frp inc,		\ adj frp (pop rp)
    frp ) fipH ld,
    frp inc,
    next,
  End-Code

  Code execute   ( xt -- ) \ execute colon definition
   'E dout,
    tos->W,
    tos pop,
    W XW ld,		\ update copy W (CFA)
    W ) W ld,		\ get addr to jmp to
    W ) jp,		\ load addr in PC
  End-Code

  Code ?branch   ( f -- ) \ jump on f<>0
    tos->W,
    W bool,		\ test tos
    tos pop,		\ get new tos
    0= IF, 		\ fip is pointing to bra offset
	fip->W,
	W ) W ld,	\ get the offset
	fip W add,	\ adjust fip (in W)
       ELSE,
	fip inc,	\ skip bra addr
	fip inc,
        W fip ex,	\ put fip in W (and W in fip)
       THEN,
    next-w,
   End-Code


\ memory access

\ its ok to use XW for scratch since it gets set again in next,

  Code @        ( addr -- n ) \ read cell
    tos->W,
    W ) W ld,
    W->tos,
    next,
   End-Code

  Code !        ( n addr -- ) \ write cell
    tos->W,
    W XW ld,	\ use XW as scratch reg for addr
    W pop,	\ n->W
    W XW ) ld,	\ n->(addr)
    tos pop,
    next,
   End-Code

\ Rabbit port access

\ internal byte i/o port data read to data stack
  Code P@	( addr -- n )
    tos->W,
    ioi,
    W ) WL ld,
    rA xor,		\ clear A
    rA WH ld,
    W->tos,
    next,
  End-Code

\ internal byte i/o port write
  Code P!	( n addr -- )
    tos->W,
    W XW ld,		\ use XW as scratch reg
    W pop,		\ get c
    ioi,
    WL XW ) ld,
    tos pop,		\ 2 things off stack
    next,
  End-Code

\ external byte i/o port data read to data stack
  Code IO@      ( addr -- n )
    tos->W,
    W XW ld,
    ioe,
    XW ) WL ld,         \ data bus is 8 bits!
    rA xor,             \ clear A
    rA WH ld,
    W->tos,
    next,
  End-Code


\ external byte i/o port write
  Code IO!      ( n addr -- )
    tos->W,
    W XW ld,            \ use XW as scratch reg
    W pop,              \ get n
    ioe,
    WL XW ) ld,         \ data bus is 8 bits!
    tos pop,            \ 2 things off stack
    next,
  End-Code 


\ data stack and return stack address

  Code sp@      ( -- sp ) \ get data stack address
    tos push,  \ make room
    0 W ld,
    fsp W add,	\ W = W + fsp = 0 + fsp = fsp
    W->tos,
    next,
   End-Code

  Code sp!      ( sp -- ) \ set data stack address
    tos->W,
    W fsp ld,
    tos pop,
    next,
  End-Code

  Code rp@      ( -- rp ) \ get return stack address
    tos push,	\ make room
    frp W ld,
    W->tos,
    next,
  End-Code

  Code rp!      ( rp -- ) \ set return stack address
    tos->W,
    W frp ld,
    tos pop,
    next,
  End-Code


\ arithmetic and logic

  Code +        ( n1 n2 -- n3 ) \ addition
    W pop,	\ n1 -> W
    tos W add,	\ W = W + tos
    W->tos,
    next,
   End-Code

  Code xor      ( n1 n2 -- n3 ) \ logic XOR
    W pop,
    tosL rA ld,
    WL xor,	\ A = A xor WL
    rA tosL ld,
    tosH rA ld,
    WH xor,	\ A = A xor WH
    rA tosH ld,
    next,
   End-Code

  Code and      ( n1 n2 -- n3 ) \ logic AND
    W pop,
    tosL rA ld,
    WL and,	\ A = A and WL
    rA tosL ld,
    tosH rA ld,
    WH and,	\ A = A and WH
    rA tosH ld,
    next,
   End-Code


\ i/o

\ if lastkey is 0 then system has processed last input query
\ (key) only gets input char (waits if not available) when lastkey=0
\ lastkey is set to 0 on every call to (key)
\ (key) returns lastkey if <>0

  Variable lastkey     \ used to sync (key) & (key?)

  Code (key)    ( -- char ) \ get character
    tos push,		\ make room
    lastkey ) rA ld,
    rA or,
    0=
    IF,
Label RxWait
    ioi,
    SERIAL STATREG + ) rA ld,
    rA 7 bit,	\ rx char ready?
    RxWait cc=Z jr,	\ wait until ready
    ioi,
    SERIAL ) rA ld,	\ get char
    THEN,
    rA tosL ld,		\ char -> tosL
    rA xor,		\ clr A	
    rA lastkey ) ld,	\ clr lastkey
    rA tosH ld,		\ clr tosH
    next,
   End-Code

  Code (emit)     ( char -- ) \ put character
Label TxChar
    ioi,
    SERIAL STATREG + ) rA ld,
    rA 2 bit,	\ tx busy?
    TxChar cc=NZ jr,	\ loop until not busy
    tosL rA ld,
    ioi,
    rA SERIAL ) ld,	\ send char
    tos pop,		\ (sp)+ -> tos
    next,
  End-Code

\ ==============================================================
\ additional words
\ ==============================================================

\ memory character access
  Code c@       ( addr -- c ) \ read character
    tos->W,
    W ) WL ld,
    rA xor,		\ clear A
    rA WH ld,
    W->tos,
    next,
   End-Code

  Code c!       ( c addr -- ) \ write character
    tos->W,
    W XW ld,		\ use XW as scratch reg for addr
    W pop,		\ get c
    WL XW ) ld,
    tos pop,		\ 2 things off stack
    next,
   End-Code


 \ moving data between stacks

  Code r>       ( -- n ; R: n -- )
    tos push,		\ make room
    frp ) W ld,
    W->tos,
    frp inc,
    frp inc,		\ post inc
    next,
   End-Code

  Code >r       ( n -- ; R: -- n )
    frp dec,		\ pre dec
    frp dec,
    tos->W,
    W frp ) ld,
    tos pop,
    next,
   End-Code

\ ==============================================================
\ useful low level words
\ ==============================================================

\ word definitions

  Code: :docon
   '1 dout,  
    tos push,		\ make room
    XW W ld,		\ current CFA restored to W
    W inc,
    W inc,
    W inc,
    W inc,		\ get addr
    W ) W ld,		\ get value
    W->tos,		\ put on tos    
    next,
  End-Code

  Code: :dodefer
   '4 dout, 
    XW W ld,		\ current CFA restored to W
    W inc,
    W inc,
    W inc,
    W inc,		\ now W is current PFA
    W ) W ld,		\ get value (deferred addr)
    W XW ld,		\ update copy of CFA
    W ) W ld,
    W ) jp,		\ go there
  End-Code


 \ branch and literal

  Code branch   ( -- ) \ unconditional branch to inline offset
    fip->W,
    W ) W ld,
    fip W add,
    W->fip,
    next,
   End-Code

  Code lit     ( -- n ) \ inline literal
    tos->W,
    W push,		\ make room in tos
    fip ) rA ld,
    fip inc,
    rA WL ld,
    fip ) rA ld,
    fip inc,
    rA WH ld,
    W->tos,
    next,
   End-Code


 \ data stack words

  Code dup      ( n -- n n )
    tos push,
    next,
   End-Code

  Code 2dup     ( d -- d d )
    W pop,
    W push,
    tos push,
    W push,
    next,
   End-Code

  Code drop     ( n -- )
    tos pop,
    next,
   End-Code

  Code 2drop    ( d -- )
    tos pop,
    tos pop,
    next,
   End-Code

  Code swap     ( n1 n2 -- n2 n1 )
    W pop,	\ n1->W
    tos push,	\ n2->(sp)
    W->tos,
    next,
   End-Code

  Code over     ( n1 n2 -- n1 n2 n1 )
    tos->W,	\ n2->W
    tos pop,	\ n1->tos
    tos push,	\ n1
    W push,	\ n2
    next,
   End-Code

  Code rot      ( n1 n2 n3 -- n2 n3 n1 )
    tos->W,	\ n3->W
    XW pop,	\ n2->XW
    tos pop,	\ n1->tos
    XW push,	\ n2
    W push,	\ n3
    next,
   End-Code

  Code -rot     ( n1 n2 n3 -- n3 n1 n2 )
    tos->W,	\ n3->W
    tos pop,	\ n2->tos
    XW pop,	\ n1->XW
    W push,	\ n3
    XW push,	\ n1
    next,
   End-Code


 \ return stack

  Code r@       ( -- n ; R: n -- n )
    tos push,
    frp ) W ld,
    W->tos,
    next,
  End-Code


 \ arithmetic

  Code -        ( n1 n2 -- n3 ) \ Subtraction
    W pop,	\ n1 -> W
    rA or,	\ CF = 0
    tos W subc,	\ W = W -tos -C
    W->tos,
    next,
   End-Code

\ r2k has hardware multiply...
\ uses these registers:
\  ' rDE Alias fip
\  ' rHL Alias W
\  ' rBC Alias tos

  Code um*      ( u1 u2 -- ud ) \ unsigned multiply
    altd,	\ SAVE fip!
    W fip ex,	\   in alternate HL
    W pop,	\ u1 -> W
    W->fip,	\ u2 already in BC (tos)
    mul,	\ HL:BC = BC * DE
    tos push,	\ n2->(sp)
    W->tos,    
    altd,	\ RESTORE fip
    W fip ex,	\   from alt HL
    next,
  End-Code

\  Code um/mod   ( ud u -- r q ) \ unsigned divide
\    next,
\   End-Code


 \ logic

  Code or       ( n1 n2 -- n3 ) \ logic OR
    W pop,	\ n1 -> W
    tosL rA ld,
    WL or,	\ A = A | WL
    rA tosL ld,
    tosH rA ld,
    WH or,	\ A = A | WH
    rA tosH ld,
    next,
   End-Code


 \ shift only single bit shifts in rabbit2000

  Code 2/       ( n1 -- n2 ) \ arithmetic shift right
     tos->W,
     tosH rl,	\ sign bit -> C
     W rr,
     W->tos,
     next,
   End-Code

\  Code lshift   ( n1 n2 -- n3 ) \ shift n1 left n2 bits
\     next,
\   End-Code

\  Code rshift   ( n1 n2 -- n3 ) \ shift n1 right n2 bits
\     next,
\   End-Code


\ compare

  Code 0=       ( n -- f ) \ Test if 0
     0 W ld,		\ W = 0
     rA or,		\ clr CF = 0
     tos W subc,	\ W = W - tos -CF (set flags)
     0 tos ld,		\ assume false. flags unchanged
     0= IF, tos dec, THEN,	\ make true
    next,
   End-Code

  Code =        ( n1 n2 -- f ) \ Test if equal
    W pop,		\ W = n1
    rA or,		\ clr CF = 0
    tos W subc,		\ W = W - tos - CF (set flags)
    0 tos ld,		\ assume false. flags unchanged
    0= IF, tos dec, THEN,
    next,
   End-Code


 \ additonal io routines

  Code (key?)     ( -- f ) \ check for read sio character
    tos push,	\ make room for result
    lastkey tos ld,	\ use tos for tmp
    tos ) rA ld,
    0 W ld,	\ assume false
    rA or,	\ set Z
    0= IF,
      ioi,
      SERIAL STATREG + ) rA ld,
      rA 7 bit,		\ char ready?
      0<> IF,	\ yes
        ioi,
        SERIAL ) rA ld,	\ get char
        rA tos ) ld,	\ char -> (lastkey)
        -1 W ld,
        THEN,
    THEN,
    W->tos,
    next,
   End-Code

  Code emit?    ( -- f ) \ check for write character to sio
    tos push,
    -1 tos ld,		\ output always possible
    next,
   End-Code


  Code bye     ( -- ) \ back to bios
\ restore state
    SAVE_SP ) rSP ld,	\ restore dyn-c sp
			\ so we know where we left the regs...
    rIX pop,
    rIY pop,
    rAF pop,
    rHL pop,
    rDE pop,
    rBC pop,		\ restore state

    0 rH ld,		\ was baud. now return value
    0 rL ld,
    ret,		\ return to bios
  End-Code

  Code: bye_f ( a - )	\ headerless so not in word list
    here 2+ ,		\ where to execute...
    tos->W,		\ get addr off data stack
			\ this is our return value
    SAVE_SP ) rSP ld,	\ restore dyn-c sp
			\ so we know where we left the regs...
    rIX pop,
    rIY pop,
    rAF pop,
\    rHL pop,		\ DON'T restore over return value (addr)
    rDE pop,		\ stack fix up to toss old HL
    rDE pop,
    rBC pop,		\ restored state

    ret,
  End-Code

  : flash ( -- )	\ exit forth & flash forth image
  here
  ['] bye_f execute 
  ;  


  : restore ( -- )	\ exit forth & restore initial image
  -1
  ['] bye_f execute
  ;


  Code: :doesjump
  End-Code

